@@ -1,10 +1,10 @@
 // Copyright 2021 Oxide Computer Company
 
 use crate::type_entry::{
-    EnumTagType, TypeEntry, TypeEntryEnum, TypeEntryStruct, Variant, VariantDetails,
+    EnumTagType, TypeEntry, TypeEntryDetails, TypeEntryEnum, TypeEntryStruct, Variant,
+    VariantDetails,
 };
-use crate::util::{all_mutually_exclusive, recase};
-use convert_case::Case;
+use crate::util::{all_mutually_exclusive, recase, Case};
 use schemars::schema::{
     ArrayValidation, InstanceType, Metadata, ObjectValidation, Schema, SchemaObject, SingleOrVec,
     SubschemaValidation,
@@ -362,28 +362,28 @@ impl TypeSpace {
         format: &Option<String>,
         _validation: &Option<Box<schemars::schema::StringValidation>>,
     ) -> Result<(TypeEntry, &'a Option<Box<Metadata>>)> {
-        trait OptionIsNoneOrDefault {
-            fn is_none_or_default(&self) -> bool;
-        }
+        // trait OptionIsNoneOrDefault {
+        //     fn is_none_or_default(&self) -> bool;
+        // }
 
-        impl<T> OptionIsNoneOrDefault for Option<T>
-        where
-            T: Default + PartialEq,
-        {
-            fn is_none_or_default(&self) -> bool {
-                match self {
-                    Some(t) => t == &T::default(),
-                    None => true,
-                }
-            }
-        }
+        // impl<T> OptionIsNoneOrDefault for Option<T>
+        // where
+        //     T: Default + PartialEq,
+        // {
+        //     fn is_none_or_default(&self) -> bool {
+        //         match self {
+        //             Some(t) => t == &T::default(),
+        //             None => true,
+        //         }
+        //     }
+        // }
         match format.as_ref().map(String::as_str) {
             None => {
                 // TODO we'll need to deal with strings with lengths and
                 // patterns, but it seems like a pain in the neck so I'm
                 // punting for now.
                 // assert!(validation.is_none_or_default(), "{:#?}", validation);
-                Ok((TypeEntry::String, metadata))
+                Ok((TypeEntryDetails::String.into(), metadata))
             }
 
             Some("uuid") => {
@@ -407,8 +407,12 @@ impl TypeSpace {
                 ))
             }
 
+            Some("ipv6") => Ok((TypeEntry::new_builtin("std::net::Ipv6Addr"), metadata)),
+
             // TODO random types I'm not sure what to do with
-            Some("uri" | "uri-template" | "email" | "ip") => Ok((TypeEntry::String, metadata)),
+            Some("uri" | "uri-template" | "email" | "ip") => {
+                Ok((TypeEntryDetails::String.into(), metadata))
+            }
 
             unhandled => todo!("{:#?}", unhandled),
         }
@@ -442,7 +446,7 @@ impl TypeSpace {
                     None
                 }
                 serde_json::Value::String(value) => {
-                    let (name, rename) = recase(value.clone(), Case::Pascal);
+                    let (name, rename) = recase(value, Case::Pascal);
                     Some(Ok(Variant {
                         name,
                         rename,
@@ -459,7 +463,8 @@ impl TypeSpace {
             EnumTagType::External,
             variants,
             false,
-        );
+        )
+        .into();
 
         if has_null {
             ty = self.type_to_option(ty);
@@ -505,6 +510,8 @@ impl TypeSpace {
             ("", "std::num::NonZeroU32", 1.0, u32::MAX as f64),
             ("uint", "u32", u32::MIN as f64, u32::MAX as f64),
             ("uint32", "u32", u32::MIN as f64, u32::MAX as f64),
+            // TODO all these are wrong as casting to an f64 loses precision.
+            // However, schemars stores everything as an f64 so... meh for now.
             ("int64", "i64", i64::MIN as f64, i64::MAX as f64),
             ("", "std::num::NonZeroU64", 1.0, u64::MAX as f64),
             ("uint64", "u64", u64::MIN as f64, u64::MAX as f64),
@@ -520,6 +527,18 @@ impl TypeSpace {
                     && (min.is_none() || min == Some(*imin))
                     && (max.is_none() || max == Some(*imax))
                 {
+                    // If there's a default value and it's either not a number
+                    // or outside of the range for this format, return an
+                    // error.
+                    if let Some(default) = metadata
+                        .as_ref()
+                        .and_then(|m| m.default.as_ref())
+                        .and_then(|v| v.as_f64())
+                    {
+                        if default < *imin || default > *imax {
+                            return Err(Error::InvalidDefaultValue);
+                        }
+                    }
                     return Ok((TypeEntry::new_integer(ty), metadata));
                 }
 
@@ -532,6 +551,23 @@ impl TypeSpace {
             }
         }
 
+        // We check the default value here since we have the min and max
+        // close at hand.
+        if let Some(default) = metadata.as_ref().and_then(|m| m.default.as_ref()) {
+            // TODO it's imprecise (in every sense of the word) to use an
+            // f64 here, but we're already constrained by the schemars
+            // representation so ... it's probably the best we can do at
+            // the moment.
+            match (default.as_f64(), min, max) {
+                (Some(_), None, None) => Some(()),
+                (Some(value), None, Some(fmax)) if value <= fmax => Some(()),
+                (Some(value), Some(fmin), None) if value >= fmin => Some(()),
+                (Some(value), Some(fmin), Some(fmax)) if value >= fmin && value <= fmax => Some(()),
+                _ => None,
+            }
+            .ok_or(Error::InvalidDefaultValue)?;
+        }
+
         // See if the value bounds fit within a known type.
         let maybe_type = match (min, max) {
             (None, Some(max)) => formats.iter().find_map(|(_, ty, _, imax)| {
@@ -564,7 +600,7 @@ impl TypeSpace {
         } else {
             // TODO we could construct a type that itself enforces the various
             // bounds.
-            // TODO failing that we should find the type that most tightly
+            // TODO failing that, we should find the type that most tightly
             // matches these bounds.
             Ok((TypeEntry::new_integer("i64"), metadata))
         }
@@ -594,7 +630,7 @@ impl TypeSpace {
         &self,
         metadata: &'a Option<Box<Metadata>>,
     ) -> Result<(TypeEntry, &'a Option<Box<Metadata>>)> {
-        Ok((TypeEntry::Unit, metadata))
+        Ok((TypeEntryDetails::Unit.into(), metadata))
     }
 
     fn convert_object<'a>(
@@ -620,22 +656,26 @@ impl TypeSpace {
                 && additional_properties.as_ref().map(AsRef::as_ref)
                     != Some(&Schema::Bool(false)) =>
             {
-                self.make_map(None, additional_properties)
+                self.make_map(type_name.into_option(), additional_properties)
             }
-            None => self.make_map(None, &None),
+            None => self.make_map(type_name.into_option(), &None),
 
             // The typical case
             Some(validation) => {
-                let tmp_type_name = get_type_name(&type_name, metadata, Case::Pascal);
+                let tmp_type_name = get_type_name(&type_name, metadata);
                 let (properties, deny_unknown_fields) =
                     self.struct_members(tmp_type_name, validation)?;
-                let ty = TypeEntryStruct::from_metadata(
-                    type_name,
-                    metadata,
-                    properties,
-                    deny_unknown_fields,
-                );
-                Ok((ty, &None))
+
+                Ok((
+                    TypeEntryStruct::from_metadata(
+                        type_name,
+                        metadata,
+                        properties,
+                        deny_unknown_fields,
+                    )
+                    .into(),
+                    &None,
+                ))
             }
         }
     }
@@ -650,8 +690,10 @@ impl TypeSpace {
             None => ref_name,
         };
         let type_id = self.ref_to_id.get(key).unwrap();
-        let ty = TypeEntry::Reference(type_id.clone());
-        Ok((ty, metadata))
+        Ok((
+            TypeEntryDetails::Reference(type_id.clone()).into(),
+            metadata,
+        ))
     }
 
     fn convert_all_of<'a>(
@@ -665,7 +707,6 @@ impl TypeSpace {
             return Ok((ty, metadata));
         }
 
-        // TODO make this look more like the other maybe clauses
         if let Some(ty) = self.maybe_all_of_subclass(type_name.clone(), metadata, subschemas) {
             return Ok((ty, metadata));
         }
@@ -700,6 +741,14 @@ impl TypeSpace {
             return Ok((ty, metadata));
         }
 
+        // Rust can emit "anyOf":[{"$ref":"#/definitions/C"},{"type":"null"}
+        // for Option, so match against that here to short circuit the check
+        // for mutual exclusivity below.
+        if let Some(option_type_entry) = self.maybe_option(type_name.clone(), metadata, subschemas)
+        {
+            return Ok((option_type_entry, metadata));
+        }
+
         // Check if this could be more precisely handled as a "one-of". This
         // occurs if each subschema is mutually exclusive i.e. so that exactly
         // one of them can match.
@@ -723,8 +772,9 @@ impl TypeSpace {
     /// several cases to consider:
     ///
     /// Options expressed as enums are uncommon since { "type": [ "null",
-    /// "xxx"], ... } is a much simpler construction. Nevertheless, an option
-    /// may be expressed as a one of with two subschemas where one is null.
+    /// "<type>"], ... } is a much simpler construction. Nevertheless, an
+    /// option may be expressed as a "one of" with two subschemas where one is
+    /// null.
     ///
     /// Externally tagged enums are comprised of either an enumerated set of
     /// string values or objects that have a single required member. The
@@ -783,7 +833,7 @@ impl TypeSpace {
             return Ok((ty, metadata));
         }
         let ty = self
-            .maybe_option_as_enum(type_name.clone(), metadata, subschemas)
+            .maybe_option(type_name.clone(), metadata, subschemas)
             .or_else(|| self.maybe_externally_tagged_enum(type_name.clone(), metadata, subschemas))
             .or_else(|| self.maybe_adjacently_tagged_enum(type_name.clone(), metadata, subschemas))
             .or_else(|| self.maybe_internally_tagged_enum(type_name.clone(), metadata, subschemas))
@@ -813,7 +863,7 @@ impl TypeSpace {
                     .map(|schema| Ok(self.id_for_schema(Name::Unknown, schema)?.0))
                     .collect::<Result<Vec<_>>>()?;
 
-                Ok((TypeEntry::Tuple(types), metadata))
+                Ok((TypeEntryDetails::Tuple(types).into(), metadata))
             }
 
             // Arrays and sets.
@@ -825,7 +875,7 @@ impl TypeSpace {
                 unique_items,
                 contains: None,
             } => {
-                let tmp_type_name = match get_type_name(&type_name, metadata, Case::Pascal) {
+                let tmp_type_name = match get_type_name(&type_name, metadata) {
                     Some(s) => Name::Suggested(format!("{}Item", s)),
                     None => Name::Unknown,
                 };
@@ -833,23 +883,22 @@ impl TypeSpace {
 
                 // If items are unique, this is a Set; otherwise it's an Array.
                 match unique_items {
-                    Some(true) => Ok((TypeEntry::Set(type_id), metadata)),
-                    _ => Ok((TypeEntry::Array(type_id), metadata)),
+                    Some(true) => Ok((TypeEntryDetails::Set(type_id).into(), metadata)),
+                    _ => Ok((TypeEntryDetails::Array(type_id).into(), metadata)),
                 }
             }
 
             _ => todo!("unhandled array validation {:#?}", validation),
         }
     }
+
     fn convert_array_of_any<'a>(
         &mut self,
         metadata: &'a Option<Box<Metadata>>,
     ) -> Result<(TypeEntry, &'a Option<Box<Metadata>>)> {
         let any = TypeEntry::new_builtin("serde_json::Value");
         let type_id = self.assign_type(any);
-        let ty = TypeEntry::Array(type_id);
-
-        Ok((ty, metadata))
+        Ok((TypeEntryDetails::Array(type_id).into(), metadata))
     }
 
     // TODO not sure if I want to deal with enum_values here, but we'll see...
@@ -858,7 +907,7 @@ impl TypeSpace {
         metadata: &'a Option<Box<Metadata>>,
         _enum_values: &Option<Vec<serde_json::Value>>,
     ) -> Result<(TypeEntry, &'a Option<Box<Metadata>>)> {
-        Ok((TypeEntry::new_integer("bool"), metadata))
+        Ok((TypeEntry::new_boolean(), metadata))
     }
 
     fn convert_permissive<'a>(
@@ -920,11 +969,17 @@ impl TypeSpace {
 mod tests {
     use std::num::{NonZeroU16, NonZeroU32, NonZeroU64, NonZeroU8};
 
-    use schemars::{schema_for, JsonSchema};
+    use schema::Schema;
+    use schemars::{
+        schema::{InstanceType, Metadata, NumberValidation, SchemaObject},
+        schema_for, JsonSchema,
+    };
+    use serde_json::json;
 
-    use crate::{validate_builtin, Name, TypeSpace};
+    use crate::{test_util::validate_output, validate_builtin, Error, Name, TypeSpace};
     use paste::paste;
 
+    #[track_caller]
     fn int_helper<T: JsonSchema>() {
         let schema = schema_for!(T);
 
@@ -1002,10 +1057,143 @@ mod tests {
         assert_eq!(type_space.iter_types().count(), 4);
     }
 
+    #[test]
+    fn test_trivial_cycle() {
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct A {
+            a: Box<A>,
+        }
+
+        validate_output::<A>();
+    }
+
+    #[test]
+    fn test_optional_trivial_cycle() {
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct A {
+            a: Option<Box<A>>,
+        }
+
+        validate_output::<A>();
+    }
+
+    #[test]
+    fn test_enum_trivial_cycles() {
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        enum A {
+            Variant0(u64),
+            Varant1 {
+                a: u64,
+                b: Vec<A>,
+                rop: Option<Box<A>>,
+            },
+            Variant2 {
+                a: Box<A>,
+            },
+            Variant3(u64, Box<A>),
+            Variant4(Option<Box<A>>, String),
+        }
+
+        validate_output::<A>();
+    }
+
+    #[test]
+    fn test_newtype_trivial_cycle() {
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct A(Box<A>);
+
+        validate_output::<A>();
+    }
+
+    #[test]
+    fn test_basic_option_flat() {
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct C {}
+
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct A {
+            a: Option<C>,
+        }
+
+        validate_output::<A>();
+    }
+
+    #[test]
+    fn test_unit_option() {
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct Foo;
+
+        #[derive(JsonSchema, Schema)]
+        #[allow(dead_code)]
+        struct Bar {
+            a: Option<Foo>,
+        }
+
+        validate_output::<Bar>();
+    }
+
     // TODO we can turn this on once we generate proper sets.
     #[ignore]
     #[test]
     fn test_set() {
         validate_builtin!(std::collections::BTreeSet<u32>);
     }
+
+    #[test]
+    fn test_low_default() {
+        let schema = SchemaObject {
+            instance_type: Some(InstanceType::Integer.into()),
+            format: Some("uint".to_string()),
+            metadata: Some(
+                Metadata {
+                    default: Some(json!(-1i32)),
+                    ..Default::default()
+                }
+                .into(),
+            ),
+            number: Some(NumberValidation::default().into()),
+            ..Default::default()
+        };
+
+        let mut type_space = TypeSpace::default();
+        match type_space.convert_schema_object(Name::Unknown, &schema) {
+            Err(Error::InvalidDefaultValue) => (),
+            _ => panic!("unexpected result"),
+        }
+    }
+
+    #[test]
+    fn test_high_default() {
+        let schema = SchemaObject {
+            instance_type: Some(InstanceType::Integer.into()),
+            metadata: Some(
+                Metadata {
+                    default: Some(json!(867_5309_u32)),
+                    ..Default::default()
+                }
+                .into(),
+            ),
+            number: Some(
+                NumberValidation {
+                    maximum: Some(256.0),
+                    ..Default::default()
+                }
+                .into(),
+            ),
+            ..Default::default()
+        };
+
+        let mut type_space = TypeSpace::default();
+        match type_space.convert_schema_object(Name::Unknown, &schema) {
+            Err(Error::InvalidDefaultValue) => (),
+            _ => panic!("unexpected result"),
+        }
+    }
 }
