@@ -1,8 +1,8 @@
 // Copyright 2021 Oxide Computer Company
 
-use std::collections::{BTreeMap, BTreeSet};
+use std::collections::{BTreeMap, BTreeSet, HashSet};
 
-use convert_case::{Case, Casing};
+use heck::{ToKebabCase, ToPascalCase};
 use proc_macro2::{Ident, TokenStream};
 use quote::{format_ident, quote};
 use schemars::schema::{
@@ -11,13 +11,15 @@ use schemars::schema::{
 
 use crate::{
     structs::output_struct_property,
-    type_entry::{EnumTagType, TypeEntryEnum, Variant, VariantDetails},
-    util::{constant_string_value, get_type_name, metadata_description, recase, schema_is_named},
-    Name, Result, TypeEntry, TypeSpace,
+    type_entry::{EnumTagType, TypeEntry, TypeEntryEnum, Variant, VariantDetails},
+    util::{
+        constant_string_value, get_type_name, metadata_description, recase, schema_is_named, Case,
+    },
+    Name, Result, TypeSpace,
 };
 
 impl TypeSpace {
-    pub(crate) fn maybe_option_as_enum(
+    pub(crate) fn maybe_option(
         &mut self,
         type_name: Name,
         metadata: &Option<Box<schemars::schema::Metadata>>,
@@ -26,8 +28,8 @@ impl TypeSpace {
         if subschemas.len() == 1 {
             return None;
         }
-        // Let's be as general as possible and consider the possibility that more
-        // than one subschema is the simple null.
+        // Let's be as general as possible and consider the possibility that
+        // more than one subschema is the simple null.
         let non_nulls = subschemas
             .iter()
             .filter(|schema| {
@@ -49,29 +51,37 @@ impl TypeSpace {
         Some(type_entry)
     }
 
-    // TODO these maybe_* functions need to not create new types until we're past
-    // that point at which they might return None.
     pub(crate) fn maybe_externally_tagged_enum(
         &mut self,
         type_name: Name,
         metadata: &Option<Box<schemars::schema::Metadata>>,
         subschemas: &[Schema],
     ) -> Option<TypeEntry> {
-        let mut deny_unknown_fields = false;
+        enum ProtoVariant<'a> {
+            Simple {
+                name: &'a str,
+                description: Option<String>,
+            },
+            Typed {
+                name: &'a str,
+                schema: &'a Schema,
+                description: Option<String>,
+            },
+        }
 
-        // TODO this should be done in two passes--the first to verify that the
-        // shape of the enum looks good; the second to actually produce the type.
-        let variants = subschemas
+        // Verify that this matches the shape of an externally tagged enum
+        // before we do any type conversion.
+        let proto_variants = subschemas
             .iter()
-            .map(|schema| -> Option<Vec<Variant>> {
+            .map(|schema| -> Option<Vec<ProtoVariant<'_>>> {
                 match schema {
-                    // It shouldn't be possible to encounter the "match anything"
-                    // schema here.
+                    // It shouldn't be possible to encounter the "match
+                    // anything" schema here.
                     Schema::Bool(true) => unreachable!(),
-                    // TODO It would be odd to see the "match nothing" schema here.
-                    // Let's abort for now, but we could implement this as a variant
-                    // that we'd never use... I guess.
-                    Schema::Bool(false) => unreachable!(),
+                    // It would be odd to see the "match nothing" schema here.
+                    // Let's abort for now, but we could implement this as a
+                    // variant that we'd never use... I guess.
+                    Schema::Bool(false) => todo!(),
 
                     // Strings must be simple enumerations.
                     Schema::Object(SchemaObject {
@@ -94,26 +104,17 @@ impl TypeSpace {
                         values
                             .iter()
                             .map(|value| {
-                                value.as_str().map(|variant_name| {
-                                    let name = variant_name.to_case(Case::Pascal);
-                                    let rename = if variant_name == name {
-                                        None
-                                    } else {
-                                        Some(variant_name.to_string())
-                                    };
-                                    Variant {
-                                        name,
-                                        rename,
-                                        description: metadata_description(metadata),
-                                        details: VariantDetails::Simple,
-                                    }
+                                value.as_str().map(|variant_name| ProtoVariant::Simple {
+                                    name: variant_name,
+                                    description: metadata_description(metadata),
                                 })
                             })
                             .collect()
                     }
 
-                    // Objects must have a single required member. The type of that
-                    // lone member determines the type associated with the variant.
+                    // Objects must have a single required member. The type of
+                    // that lone member determines the type associated with the
+                    // variant.
                     Schema::Object(SchemaObject {
                         metadata,
                         instance_type: Some(SingleOrVec::Single(single)),
@@ -128,46 +129,30 @@ impl TypeSpace {
                         reference: None,
                         extensions: _,
                     }) if single.as_ref() == &InstanceType::Object => {
-                        let validation = validation.as_ref();
                         if let ObjectValidation {
                             max_properties: None,
                             min_properties: None,
                             required,
                             properties,
                             pattern_properties,
-                            additional_properties: Some(additional_properties),
+                            additional_properties: _,
                             property_names: None,
-                        } = validation
+                        } = validation.as_ref()
                         {
                             if required.len() == 1
                                 && properties.len() == 1
                                 && pattern_properties.is_empty()
-                                && additional_properties.as_ref() == &Schema::Bool(false)
                             {
                                 let (prop_name, prop_type) = properties.iter().next().unwrap();
-
-                                let name = prop_name.to_case(Case::Pascal);
-                                let rename = if *prop_name == name {
-                                    None
-                                } else {
-                                    Some(prop_name.clone())
-                                };
-
                                 // If required and properties both have length 1
                                 // then this must be true for a well-constructed
                                 // schema.
                                 assert!(required.contains(prop_name));
-                                // TODO should I be doing something different with the error below?
-                                let (details, deny) = self
-                                    .external_variant(type_name.clone(), prop_name, prop_type)
-                                    .ok()?;
-                                deny_unknown_fields |= deny;
 
-                                Some(vec![Variant {
-                                    name,
-                                    rename,
+                                Some(vec![ProtoVariant::Typed {
+                                    name: prop_name,
+                                    schema: prop_type,
                                     description: metadata_description(metadata),
-                                    details,
                                 }])
                             } else {
                                 None
@@ -183,9 +168,59 @@ impl TypeSpace {
                 Some(v) => v.into_iter().map(Some).collect::<Vec<_>>(),
                 None => vec![None],
             })
-            .collect::<Option<Vec<_>>>();
+            .collect::<Option<Vec<_>>>()?;
 
-        variants.map(|variants| {
+        let variant_names = proto_variants
+            .iter()
+            .map(|proto| match proto {
+                ProtoVariant::Simple { name, .. } | ProtoVariant::Typed { name, .. } => name,
+            })
+            .collect::<HashSet<_>>();
+
+        // We can't have duplicate names in an enum.
+        if variant_names.len() != proto_variants.len() {
+            return None;
+        }
+
+        let mut deny_unknown_fields = false;
+        let variants = proto_variants
+            .into_iter()
+            .map(|proto| match proto {
+                ProtoVariant::Simple {
+                    name: variant_name,
+                    description,
+                } => {
+                    let (name, rename) = recase(variant_name, Case::Pascal);
+                    Some(Variant {
+                        name,
+                        rename,
+                        description,
+                        details: VariantDetails::Simple,
+                    })
+                }
+
+                ProtoVariant::Typed {
+                    name: variant_name,
+                    schema,
+                    description,
+                } => {
+                    let (details, deny) = self
+                        .external_variant(type_name.clone(), variant_name, schema)
+                        .ok()?;
+                    deny_unknown_fields |= deny;
+
+                    let (name, rename) = recase(variant_name, Case::Pascal);
+                    Some(Variant {
+                        name,
+                        rename,
+                        description,
+                        details,
+                    })
+                }
+            })
+            .collect::<Option<Vec<_>>>()?;
+
+        Some(
             TypeEntryEnum::from_metadata(
                 type_name,
                 metadata,
@@ -193,7 +228,8 @@ impl TypeSpace {
                 variants,
                 deny_unknown_fields,
             )
-        })
+            .into(),
+        )
     }
 
     fn external_variant(
@@ -204,7 +240,7 @@ impl TypeSpace {
     ) -> Result<(VariantDetails, bool)> {
         let prop_type_name = match type_name {
             Name::Required(name) | Name::Suggested(name) => {
-                Name::Suggested(format!("{}{}", name, variant_name.to_case(Case::Pascal)))
+                Name::Suggested(format!("{}{}", name, variant_name.to_pascal_case()))
             }
             Name::Unknown => Name::Unknown,
         };
@@ -342,7 +378,10 @@ impl TypeSpace {
                         .filter_map(|(prop_name, prop_type)| {
                             constant_string_value(prop_type).map(|value| {
                                 // Tuple with the name and a set with a single value
-                                (prop_name.clone(), [value].iter().cloned().collect())
+                                (
+                                    prop_name.clone(),
+                                    [value.to_string()].iter().cloned().collect(),
+                                )
                             })
                         })
                         .collect()
@@ -401,20 +440,23 @@ impl TypeSpace {
             .collect::<Result<Vec<_>>>()
             .ok()?;
 
-        Some(TypeEntryEnum::from_metadata(
-            type_name,
-            metadata,
-            EnumTagType::Internal { tag: tag.clone() },
-            variants,
-            deny_unknown_fields,
-        ))
+        Some(
+            TypeEntryEnum::from_metadata(
+                type_name,
+                metadata,
+                EnumTagType::Internal { tag: tag.clone() },
+                variants,
+                deny_unknown_fields,
+            )
+            .into(),
+        )
     }
 
     fn internal_variant(&mut self, validation: &ObjectValidation, tag: &str) -> Result<Variant> {
         if validation.properties.len() == 1 {
             let (tag_name, schema) = validation.properties.iter().next().unwrap();
             let variant_name = constant_string_value(schema).unwrap();
-            let (name, rename) = recase(variant_name, Case::Pascal);
+            let (name, rename) = recase(&variant_name, Case::Pascal);
 
             // The lone property must be our tag.
             assert_eq!(tag_name, tag);
@@ -430,7 +472,7 @@ impl TypeSpace {
         } else {
             let tag_schema = validation.properties.get(tag).unwrap();
             let variant_name = constant_string_value(tag_schema).unwrap();
-            let (name, rename) = recase(variant_name, Case::Pascal);
+            let (name, rename) = recase(&variant_name, Case::Pascal);
 
             // Make a new object validation that omits the tag.
             let mut new_validation = validation.clone();
@@ -528,13 +570,16 @@ impl TypeSpace {
             .collect::<Result<Vec<_>>>()
             .ok()?;
 
-        Some(TypeEntryEnum::from_metadata(
-            type_name,
-            metadata,
-            EnumTagType::Adjacent { tag, content },
-            variants,
-            deny_unknown_fields,
-        ))
+        Some(
+            TypeEntryEnum::from_metadata(
+                type_name,
+                metadata,
+                EnumTagType::Adjacent { tag, content },
+                variants,
+                deny_unknown_fields,
+            )
+            .into(),
+        )
     }
 
     fn adjacent_variant(
@@ -612,7 +657,7 @@ impl TypeSpace {
         metadata: &Option<Box<schemars::schema::Metadata>>,
         subschemas: &[Schema],
     ) -> Result<TypeEntry> {
-        let tmp_type_name = get_type_name(&type_name, metadata, Case::Pascal);
+        let tmp_type_name = get_type_name(&type_name, metadata);
 
         let mut names_from_variants = true;
         let mut common_prefix = None;
@@ -678,7 +723,8 @@ impl TypeSpace {
             EnumTagType::Untagged,
             variants,
             deny_unknown_fields,
-        ))
+        )
+        .into())
     }
 }
 
@@ -701,14 +747,12 @@ fn get_object(schema: &Schema) -> Option<(Option<&Metadata>, &ObjectValidation)>
             reference: None,
             extensions: _,
         }) if single.as_ref() == &InstanceType::Object
-            && schema_none_or_false(&validation.additional_properties) =>
+            && schema_none_or_false(&validation.additional_properties)
+            && validation.max_properties.is_none()
+            && validation.min_properties.is_none()
+            && validation.pattern_properties.is_empty()
+            && validation.property_names.is_none() =>
         {
-            // These are the fields we don't currently handle
-            assert!(validation.max_properties.is_none());
-            assert!(validation.min_properties.is_none());
-            assert!(validation.pattern_properties.is_empty());
-            assert!(validation.property_names.is_none());
-
             Some((metadata.as_ref().map(|m| m.as_ref()), validation.as_ref()))
         }
 
@@ -717,9 +761,8 @@ fn get_object(schema: &Schema) -> Option<(Option<&Metadata>, &ObjectValidation)>
     }
 }
 
-// TODO: https://github.com/GREsau/schemars/pull/99
-// Really this should be Schema::Bool(false), but additional_properties is used
-// inconsistently... in schemars, but also generally.
+// We infer from a Some(Schema::Bool(false)) or None value that either nothing
+// or nothing of importance is in the additional properties.
 fn schema_none_or_false(additional_properties: &Option<Box<Schema>>) -> bool {
     matches!(
         additional_properties.as_ref().map(Box::as_ref),
@@ -730,17 +773,24 @@ fn schema_none_or_false(additional_properties: &Option<Box<Schema>>) -> bool {
 /// Get the string that represents the common prefix, considering only
 /// case-relevant boundaries.
 fn get_common_prefix(name: &str, prefix: &str) -> String {
-    name.to_case(Case::Kebab)
+    // Convert both to kebab case, split by '-' and zip them together. Pick up
+    // components in order while they're equal, join, and convert to Pascal
+    // case for a variant name.
+    name.to_kebab_case()
         .split('-')
-        .zip(prefix.to_case(Case::Kebab).split('-'))
+        .zip(prefix.to_kebab_case().split('-'))
         .take_while(|(a, b)| a == b)
         .map(|(a, _)| a)
         .collect::<Vec<&str>>()
         .join("-")
-        .to_case(Case::Pascal)
+        .to_pascal_case()
 }
 
-pub(crate) fn output_variant(variant: &Variant, type_space: &TypeSpace) -> TokenStream {
+pub(crate) fn output_variant(
+    variant: &Variant,
+    type_space: &TypeSpace,
+    type_name: &str,
+) -> TokenStream {
     let name = format_ident!("{}", variant.name);
     let doc = match &variant.description {
         Some(s) => quote! {#[doc = #s]},
@@ -774,16 +824,25 @@ pub(crate) fn output_variant(variant: &Variant, type_space: &TypeSpace) -> Token
         }
 
         VariantDetails::Struct(props) => {
-            let properties = props
+            let (prop_streams, prop_defaults): (Vec<_>, Vec<_>) = props
                 .iter()
-                .map(|prop| output_struct_property(prop, type_space, false))
-                .collect::<Vec<_>>();
+                .map(|prop| {
+                    output_struct_property(
+                        prop,
+                        type_space,
+                        false,
+                        &format!("{}{}", type_name, &variant.name),
+                    )
+                })
+                .unzip();
             quote! {
                 #doc
                 #rename
                 #name {
-                    #(#properties)*
+                    #(#prop_streams)*
                 },
+
+                #(#prop_defaults)*
             }
         }
     }
@@ -817,7 +876,7 @@ pub(crate) fn enum_impl(type_name: &Ident, variants: &[Variant]) -> TokenStream
             quote! {
                 impl ToString for #type_name {
                     fn to_string(&self) -> String {
-                        match self {
+                        match *self {
                             #(#match_variants),*
                         }
                     }
@@ -834,6 +893,7 @@ pub(crate) fn enum_impl(type_name: &Ident, variants: &[Variant]) -> TokenStream
 mod tests {
     use std::collections::HashSet;
 
+    use quote::quote;
     use schema::Schema;
     use schemars::{
         schema::{InstanceType, RootSchema, SchemaObject, SingleOrVec},
@@ -844,7 +904,7 @@ mod tests {
     use crate::{
         test_util::{validate_output, validate_output_for_untagged_enm},
         type_entry::{EnumTagType, TypeEntryEnum, Variant, VariantDetails},
-        Name, TypeEntry, TypeId, TypeSpace,
+        Name, TypeEntryDetails, TypeId, TypeSpace,
     };
 
     #[allow(dead_code)]
@@ -1015,11 +1075,12 @@ mod tests {
             )
             .unwrap();
 
-        match ty {
-            TypeEntry::Enum(TypeEntryEnum {
+        match &ty.details {
+            TypeEntryDetails::Enum(TypeEntryEnum {
                 name,
                 rename: None,
                 description: None,
+                default: None,
                 tag_type: EnumTagType::Untagged,
                 variants,
                 deny_unknown_fields: _,
@@ -1085,8 +1146,8 @@ mod tests {
         // This confirms in particular that the tag type is untagged and
         // therefore that the other enum tagging regimes did not match.
         assert!(matches!(
-            ty,
-            TypeEntry::Enum(TypeEntryEnum {
+            &ty.details,
+            TypeEntryDetails::Enum(TypeEntryEnum {
                 rename: None,
                 description: None,
                 tag_type: EnumTagType::Untagged,
@@ -1175,12 +1236,12 @@ mod tests {
             .convert_schema_object(Name::Unknown, &schema.schema)
             .unwrap();
 
-        if let TypeEntry::Enum(TypeEntryEnum {
+        if let TypeEntryDetails::Enum(TypeEntryEnum {
             variants,
             tag_type,
             deny_unknown_fields,
             ..
-        }) = &type_entry
+        }) = &type_entry.details
         {
             let variant_names = variants
                 .iter()
@@ -1195,7 +1256,7 @@ mod tests {
     }
 
     #[test]
-    fn test_maybe_option_as_enum() {
+    fn test_maybe_option() {
         let subschemas = vec![
             SchemaObject {
                 instance_type: Some(SingleOrVec::Single(Box::new(InstanceType::String))),
@@ -1211,10 +1272,10 @@ mod tests {
 
         let mut type_space = TypeSpace::default();
         let type_entry = type_space
-            .maybe_option_as_enum(Name::Unknown, &None, &subschemas)
+            .maybe_option(Name::Unknown, &None, &subschemas)
             .unwrap();
 
-        assert_eq!(type_entry, TypeEntry::Option(TypeId(1)))
+        assert_eq!(type_entry.details, TypeEntryDetails::Option(TypeId(1)))
     }
 
     #[test]
@@ -1290,8 +1351,8 @@ mod tests {
         let type_id = type_space.ref_to_id.get("workflow-step").unwrap();
         let type_entry = type_space.id_to_entry.get(type_id).unwrap();
 
-        match &type_entry {
-            TypeEntry::Enum(TypeEntryEnum {
+        match &type_entry.details {
+            TypeEntryDetails::Enum(TypeEntryEnum {
                 tag_type,
                 variants,
                 deny_unknown_fields: _,
@@ -1314,9 +1375,78 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_untagged_enum_nice_names() {
+        let schema_json = r##"
+        {
+            "definitions": {
+                "IpNet": {
+                    "oneOf": [
+                        {
+                            "title": "V4",
+                            "allOf": [
+                                {
+                                    "$ref": "#/components/schemas/Ipv4Net"
+                                }
+                            ]
+                        },
+                        {
+                            "title": "V6",
+                            "allOf": [
+                                {
+                                    "$ref": "#/components/schemas/Ipv4Net"
+                                }
+                            ]
+                        }
+                    ]
+                },
+                "Ipv4Net": {
+                    "type": "string"
+                },
+                "Ipv6Net": {
+                    "type": "string"
+                }
+            }
+        }
+        "##;
+
+        let schema: RootSchema = serde_json::from_str(schema_json).unwrap();
+
+        let mut type_space = TypeSpace::default();
+        type_space.add_ref_types(schema.definitions).unwrap();
+
+        let type_id = type_space.ref_to_id.get("IpNet").unwrap();
+        let type_entry = type_space.id_to_entry.get(type_id).unwrap();
+
+        match &type_entry.details {
+            TypeEntryDetails::Enum(TypeEntryEnum {
+                tag_type,
+                variants,
+                deny_unknown_fields: _,
+                ..
+            }) => {
+                assert_eq!(tag_type, &EnumTagType::Untagged);
+                //assert_eq!(deny_unknown_fields, &true);
+                let variant_names = variants
+                    .iter()
+                    .map(|variant| variant.name.clone())
+                    .collect::<HashSet<_>>();
+                assert_eq!(
+                    variant_names,
+                    ["V4", "V6"]
+                        .iter()
+                        .map(ToString::to_string)
+                        .collect::<HashSet<_>>()
+                );
+            }
+            _ => panic!("{:#?}", type_entry),
+        }
+    }
+
     #[allow(dead_code)]
     #[derive(Serialize, JsonSchema, Schema)]
-    // TODO change this to deny_unknown_fields, but there's a bug in schemars
+    // TODO change this to deny_unknown_fields, but there's a bug in schemars;
+    // see https://github.com/GREsau/schemars/pull/113
     #[serde(tag = "tag")]
     enum InternalSimple {
         Shadrach,
@@ -1328,4 +1458,46 @@ mod tests {
     fn test_internal_deny_simple() {
         validate_output::<InternalSimple>();
     }
+
+    #[test]
+    fn test_result() {
+        let mut type_space = TypeSpace::default();
+        let schema = schema_for!(Result<u32, String>);
+        let subschemas = schema.schema.subschemas.unwrap().one_of.unwrap();
+        let type_entry = type_space
+            .maybe_externally_tagged_enum(Name::Required("ResultX".to_string()), &None, &subschemas)
+            .unwrap();
+        let actual = type_entry.output(&type_space);
+        let expected = quote! {
+            #[derive(Serialize, Deserialize, Debug, Clone)]
+            pub enum ResultX {
+                Ok(u32),
+                Err(String),
+            }
+        };
+        assert_eq!(actual.to_string(), expected.to_string());
+    }
+
+    #[test]
+    fn test_result_derives() {
+        let mut type_space = TypeSpace::default();
+        type_space.add_derive(quote! { A });
+        type_space.add_derive(quote! { B });
+        type_space.add_derive(quote! { C });
+        type_space.add_derive(quote! { D });
+        let schema = schema_for!(Result<u32, String>);
+        let subschemas = schema.schema.subschemas.unwrap().one_of.unwrap();
+        let type_entry = type_space
+            .maybe_externally_tagged_enum(Name::Required("ResultX".to_string()), &None, &subschemas)
+            .unwrap();
+        let actual = type_entry.output(&type_space);
+        let expected = quote! {
+            #[derive(Serialize, Deserialize, Debug, Clone, A, B, C, D)]
+            pub enum ResultX {
+                Ok(u32),
+                Err(String),
+            }
+        };
+        assert_eq!(actual.to_string(), expected.to_string());
+    }
 }
