@@ -1,7 +1,8 @@
 // Copyright 2021 Oxide Computer Company
 
-use convert_case::Case;
-use proc_macro2::TokenStream;
+use std::collections::BTreeSet;
+
+use proc_macro2::{Punct, Spacing, TokenStream, TokenTree};
 use quote::{format_ident, quote, ToTokens};
 use schemars::schema::Metadata;
 
@@ -9,7 +10,7 @@ use crate::{
     enums::{enum_impl, output_variant},
     structs::output_struct_property,
     util::{get_type_name, metadata_description},
-    Name, TypeId, TypeSpace,
+    DefaultImpl, Name, TypeId, TypeSpace,
 };
 
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
@@ -17,6 +18,7 @@ pub(crate) struct TypeEntryEnum {
     pub name: String,
     pub rename: Option<String>,
     pub description: Option<String>,
+    pub default: Option<DefaultValue>,
     pub tag_type: EnumTagType,
     pub variants: Vec<Variant>,
     pub deny_unknown_fields: bool,
@@ -27,6 +29,7 @@ pub(crate) struct TypeEntryStruct {
     pub name: String,
     pub rename: Option<String>,
     pub description: Option<String>,
+    pub default: Option<DefaultValue>,
     pub properties: Vec<StructProperty>,
     pub deny_unknown_fields: bool,
 }
@@ -36,26 +39,55 @@ pub(crate) struct TypeEntryNewtype {
     pub name: String,
     pub rename: Option<String>,
     pub description: Option<String>,
+    pub default: Option<DefaultValue>,
     pub type_id: TypeId,
 }
 
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub(crate) struct DefaultValue(pub serde_json::Value);
+impl DefaultValue {
+    pub(crate) fn new(value: serde_json::Value) -> Self {
+        Self(value)
+    }
+}
+
+impl Ord for DefaultValue {
+    fn cmp(&self, _: &Self) -> std::cmp::Ordering {
+        std::cmp::Ordering::Equal
+    }
+}
+impl PartialOrd for DefaultValue {
+    fn partial_cmp(&self, _: &Self) -> Option<std::cmp::Ordering> {
+        Some(std::cmp::Ordering::Equal)
+    }
+}
+
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub(crate) struct TypeEntry {
+    pub details: TypeEntryDetails,
+    pub derives: Option<BTreeSet<&'static str>>,
+}
+
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
-pub(crate) enum TypeEntry {
+pub(crate) enum TypeEntryDetails {
     Enum(TypeEntryEnum),
     Struct(TypeEntryStruct),
     Newtype(TypeEntryNewtype),
 
     Option(TypeId),
+    Box(TypeId),
     Array(TypeId),
-    Map(TypeId, TypeId),
+    Map(TypeId),
     Set(TypeId),
     Tuple(Vec<TypeId>),
     Unit,
     /// Built-in complex types with no type generics such as Uuid
     BuiltIn(String),
-    /// Integers and booleans
-    Integral(String),
-    /// Floating point numbers; critical values that cannot be Eq
+    /// Boolean
+    Boolean,
+    /// Integers
+    Integer(String),
+    /// Floating point numbers; not Eq, Ord, or Hash
     Float(String),
     /// Strings... which we handle a little specially.
     String,
@@ -92,23 +124,31 @@ pub(crate) enum VariantDetails {
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
 pub(crate) struct StructProperty {
     pub name: String,
-    pub serde_naming: SerdeNaming,
-    pub serde_rules: SerdeRules,
+    pub rename: StructPropertyRename,
+    pub state: StructPropertyState,
     pub description: Option<String>,
     pub type_id: TypeId,
 }
 
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
-pub(crate) enum SerdeNaming {
+pub(crate) enum StructPropertyRename {
     None,
     Rename(String),
     Flatten,
 }
 
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
-pub(crate) enum SerdeRules {
-    None,
+pub(crate) enum StructPropertyState {
+    Required,
     Optional,
+    Default(DefaultValue),
+}
+
+#[derive(Debug)]
+pub(crate) enum DefaultKind {
+    Intrinsic,
+    Specific,
+    Generic(DefaultImpl),
 }
 
 impl TypeEntryEnum {
@@ -118,15 +158,16 @@ impl TypeEntryEnum {
         tag_type: EnumTagType,
         variants: Vec<Variant>,
         deny_unknown_fields: bool,
-    ) -> TypeEntry {
-        let name = get_type_name(&type_name, metadata, Case::Pascal).unwrap();
+    ) -> TypeEntryDetails {
+        let name = get_type_name(&type_name, metadata).unwrap();
         let rename = None;
         let description = metadata_description(metadata);
 
-        TypeEntry::Enum(Self {
+        TypeEntryDetails::Enum(Self {
             name,
             rename,
             description,
+            default: None,
             tag_type,
             variants,
             deny_unknown_fields,
@@ -140,15 +181,21 @@ impl TypeEntryStruct {
         metadata: &Option<Box<Metadata>>,
         properties: Vec<StructProperty>,
         deny_unknown_fields: bool,
-    ) -> TypeEntry {
-        let name = get_type_name(&type_name, metadata, Case::Pascal).unwrap();
+    ) -> TypeEntryDetails {
+        let name = get_type_name(&type_name, metadata).unwrap();
         let rename = None;
         let description = metadata_description(metadata);
+        let default = metadata
+            .as_ref()
+            .and_then(|m| m.default.as_ref())
+            .cloned()
+            .map(DefaultValue::new);
 
-        TypeEntry::Struct(Self {
+        TypeEntryDetails::Struct(Self {
             name,
             rename,
             description,
+            default,
             properties,
             deny_unknown_fields,
         })
@@ -156,40 +203,65 @@ impl TypeEntryStruct {
 }
 
 impl TypeEntryNewtype {
-    pub(crate) fn from_metadata(
+    pub(crate) fn from_metadata_with_default(
         type_name: Name,
         metadata: &Option<Box<Metadata>>,
         type_id: TypeId,
-    ) -> TypeEntry {
-        let name = get_type_name(&type_name, metadata, Case::Pascal).unwrap();
+    ) -> TypeEntryDetails {
+        let name = get_type_name(&type_name, metadata).unwrap();
         let rename = None;
         let description = metadata_description(metadata);
 
-        TypeEntry::Newtype(Self {
+        TypeEntryDetails::Newtype(Self {
             name,
             rename,
             description,
+            default: None,
             type_id,
         })
     }
 }
 
+impl From<TypeEntryDetails> for TypeEntry {
+    fn from(details: TypeEntryDetails) -> Self {
+        Self {
+            details,
+            derives: None,
+        }
+    }
+}
+
 impl TypeEntry {
     pub(crate) fn new_builtin<S: ToString>(type_name: S) -> Self {
-        TypeEntry::BuiltIn(type_name.to_string())
+        TypeEntry {
+            details: TypeEntryDetails::BuiltIn(type_name.to_string()),
+            derives: None,
+        }
+    }
+    pub(crate) fn new_boolean() -> Self {
+        TypeEntry {
+            details: TypeEntryDetails::Boolean,
+            derives: None,
+        }
     }
     pub(crate) fn new_integer<S: ToString>(type_name: S) -> Self {
-        TypeEntry::Integral(type_name.to_string())
+        TypeEntry {
+            details: TypeEntryDetails::Integer(type_name.to_string()),
+            derives: None,
+        }
     }
     pub(crate) fn new_float<S: ToString>(type_name: S) -> Self {
-        TypeEntry::Float(type_name.to_string())
+        TypeEntry {
+            details: TypeEntryDetails::Float(type_name.to_string()),
+            derives: None,
+        }
     }
 
     pub(crate) fn name(&self) -> Option<&String> {
-        match self {
-            TypeEntry::Enum(TypeEntryEnum { name, .. })
-            | TypeEntry::Struct(TypeEntryStruct { name, .. })
-            | TypeEntry::Newtype(TypeEntryNewtype { name, .. }) => Some(name),
+        match &self.details {
+            TypeEntryDetails::Enum(TypeEntryEnum { name, .. })
+            | TypeEntryDetails::Struct(TypeEntryStruct { name, .. })
+            | TypeEntryDetails::Newtype(TypeEntryNewtype { name, .. }) => Some(name),
 
             _ => None,
         }
@@ -197,30 +269,34 @@ impl TypeEntry {
 
     pub(crate) fn output(&self, type_space: &TypeSpace) -> TokenStream {
         let mut derives = vec![
-            quote! {Serialize},
-            quote! {Deserialize},
-            quote! {Debug},
-            quote! {Clone},
+            quote! { Serialize },
+            quote! { Deserialize },
+            quote! { Debug },
+            quote! { Clone },
         ];
 
-        match self {
-            TypeEntry::Enum(TypeEntryEnum {
+        derives.extend(type_space.extra_derives.clone());
+
+        match &self.details {
+            TypeEntryDetails::Enum(TypeEntryEnum {
                 name,
                 rename,
                 description,
+                default,
                 tag_type,
                 variants,
                 deny_unknown_fields,
             }) => {
                 let doc = description.as_ref().map(|desc| quote! { #[doc = #desc] });
 
-                // TODO this is a one-off for Eq
+                // TODO this is a one-off for some useful traits
                 if variants
                     .iter()
                     .all(|variant| matches!(variant.details, VariantDetails::Simple))
                 {
                     derives.extend(
                         vec![
+                            quote! { Copy },
                             quote! { PartialOrd },
                             quote! { Ord },
                             quote! { PartialEq },
@@ -261,11 +337,22 @@ impl TypeEntry {
 
                 let variants_decl = variants
                     .iter()
-                    .map(|variant| output_variant(variant, type_space))
+                    .map(|variant| output_variant(variant, type_space, name))
                     .collect::<Vec<_>>();
 
                 let enum_impl = enum_impl(&type_name, variants);
 
+                let default_impl = default.as_ref().map(|value| {
+                    let default_stream = self.value(type_space, &value.0).unwrap();
+                    quote! {
+                        impl Default for #type_name {
+                            fn default() -> Self {
+                                #default_stream
+                            }
+                        }
+                    }
+                });
+
                 quote! {
                     #doc
                     #[derive(#(#derives),*)]
@@ -275,13 +362,15 @@ impl TypeEntry {
                     }
 
                     #enum_impl
+                    #default_impl
                 }
             }
 
-            TypeEntry::Struct(TypeEntryStruct {
+            TypeEntryDetails::Struct(TypeEntryStruct {
                 name,
                 rename,
                 description,
+                default,
                 properties,
                 deny_unknown_fields,
             }) => {
@@ -301,25 +390,41 @@ impl TypeEntry {
                 };
 
                 let type_name = format_ident!("{}", name);
-                let properties = properties
+                let (prop_streams, prop_defaults): (Vec<_>, Vec<_>) = properties
                     .iter()
-                    .map(|prop| output_struct_property(prop, type_space, true))
-                    .collect::<Vec<_>>();
+                    .map(|prop| output_struct_property(prop, type_space, true, name))
+                    .unzip();
+
+                let default_impl = default.as_ref().map(|value| {
+                    let default_stream = self.value(type_space, &value.0).unwrap();
+                    quote! {
+                        impl Default for #type_name {
+                            fn default() -> Self {
+                                #default_stream
+                            }
+                        }
+                    }
+                });
 
                 quote! {
                     #doc
                     #[derive(#(#derives),*)]
                     #serde
                     pub struct #type_name {
-                        #(#properties)*
+                        #(#prop_streams)*
                     }
+
+                    #(#prop_defaults)*
+
+                    #default_impl
                 }
             }
 
-            TypeEntry::Newtype(TypeEntryNewtype {
+            TypeEntryDetails::Newtype(TypeEntryNewtype {
                 name,
                 rename,
                 description,
+                default,
                 type_id,
             }) => {
                 let doc = description.as_ref().map(|desc| quote! { #[doc = #desc] });
@@ -334,6 +439,17 @@ impl TypeEntry {
                 let sub_type = type_space.id_to_entry.get(type_id).unwrap();
                 let sub_type_name = sub_type.type_ident(type_space, false);
 
+                let default_impl = default.as_ref().map(|value| {
+                    let default_stream = self.value(type_space, &value.0).unwrap();
+                    quote! {
+                        impl Default for #type_name {
+                            fn default() -> Self {
+                                #default_stream
+                            }
+                        }
+                    }
+                });
+
                 quote! {
                     #doc
                     #[derive(#(#derives),*)]
@@ -346,24 +462,28 @@ impl TypeEntry {
                             &self.0
                         }
                     }
+
+                    #default_impl
                 }
             }
 
             // These types require no definition as they're already defined.
-            TypeEntry::BuiltIn(_)
-            | TypeEntry::Integral(_)
-            | TypeEntry::Float(_)
-            | TypeEntry::String
-            | TypeEntry::Option(_)
-            | TypeEntry::Array(_)
-            | TypeEntry::Map(_, _)
-            | TypeEntry::Set(_)
-            | TypeEntry::Unit
-            | TypeEntry::Tuple(_) => quote! {},
+            TypeEntryDetails::BuiltIn(_)
+            | TypeEntryDetails::Boolean
+            | TypeEntryDetails::Integer(_)
+            | TypeEntryDetails::Float(_)
+            | TypeEntryDetails::String
+            | TypeEntryDetails::Option(_)
+            | TypeEntryDetails::Array(_)
+            | TypeEntryDetails::Map(_)
+            | TypeEntryDetails::Set(_)
+            | TypeEntryDetails::Unit
+            | TypeEntryDetails::Box(_)
+            | TypeEntryDetails::Tuple(_) => quote! {},
 
             // We should never get here as reference types should only be used
             // in-flight, but never recorded into the type space.
-            TypeEntry::Reference(_) => unreachable!(),
+            TypeEntryDetails::Reference(_) => unreachable!(),
         }
     }
 
@@ -372,23 +492,25 @@ impl TypeEntry {
     }
 
     pub(crate) fn type_ident(&self, type_space: &TypeSpace, external: bool) -> TokenStream {
-        match self {
+        match &self.details {
             // Named types.
-            TypeEntry::Enum(TypeEntryEnum { name, .. })
-            | TypeEntry::Struct(TypeEntryStruct { name, .. })
-            | TypeEntry::Newtype(TypeEntryNewtype { name, .. }) => match &type_space.type_mod {
-                Some(type_mod) if external => {
-                    let type_mod = format_ident!("{}", type_mod);
-                    let type_name = format_ident!("{}", name);
-                    quote! { #type_mod :: #type_name }
+            TypeEntryDetails::Enum(TypeEntryEnum { name, .. })
+            | TypeEntryDetails::Struct(TypeEntryStruct { name, .. })
+            | TypeEntryDetails::Newtype(TypeEntryNewtype { name, .. }) => {
+                match &type_space.type_mod {
+                    Some(type_mod) if external => {
+                        let type_mod = format_ident!("{}", type_mod);
+                        let type_name = format_ident!("{}", name);
+                        quote! { #type_mod :: #type_name }
+                    }
+                    _ => {
+                        let type_name = format_ident!("{}", name);
+                        quote! { #type_name }
+                    }
                 }
-                _ => {
-                    let type_name = format_ident!("{}", name);
-                    quote! { #type_name }
-                }
-            },
+            }
 
-            TypeEntry::Option(id) => {
+            TypeEntryDetails::Option(id) => {
                 let inner_ty = type_space
                     .id_to_entry
                     .get(id)
@@ -397,13 +519,24 @@ impl TypeEntry {
 
                 // Flatten nested Option types. This would only happen if the
                 // schema encoded it; it's an odd construction.
-                match inner_ty {
-                    TypeEntry::Option(_) => inner_ident,
+                match &inner_ty.details {
+                    TypeEntryDetails::Option(_) => inner_ident,
                     _ => quote! { Option<#inner_ident> },
                 }
             }
 
-            TypeEntry::Array(id) => {
+            TypeEntryDetails::Box(id) => {
+                let inner_ty = type_space
+                    .id_to_entry
+                    .get(id)
+                    .expect("unresolved type id for box");
+
+                let item = inner_ty.type_ident(type_space, external);
+
+                quote! { Box<#item> }
+            }
+
+            TypeEntryDetails::Array(id) => {
                 let inner_ty = type_space
                     .id_to_entry
                     .get(id)
@@ -413,33 +546,27 @@ impl TypeEntry {
                 quote! { Vec<#item> }
             }
 
-            TypeEntry::Map(key_id, value_id) => {
-                let key_ty = type_space
+            TypeEntryDetails::Map(type_id) => {
+                let inner_ty = type_space
                     .id_to_entry
-                    .get(key_id)
-                    .expect("unresolved type id for map")
-                    .type_ident(type_space, external);
-                let value_ty = type_space
-                    .id_to_entry
-                    .get(value_id)
+                    .get(type_id)
                     .expect("unresolved type id for map")
                     .type_ident(type_space, external);
 
-                quote! { std::collections::HashMap<#key_ty, #value_ty> }
+                quote! { std::collections::HashMap<String, #inner_ty> }
             }
 
-            TypeEntry::Set(id) => {
+            TypeEntryDetails::Set(id) => {
                 let inner_ty = type_space
                     .id_to_entry
                     .get(id)
                     .expect("unresolved type id for set");
                 let item = inner_ty.type_ident(type_space, external);
                 // TODO we'll want this to be a Set of some kind, but we need to get the derives right first.
-                // quote! { std::collections::BTreeSet<#item> }
                 quote! { Vec<#item> }
             }
 
-            TypeEntry::Tuple(items) => {
+            TypeEntryDetails::Tuple(items) => {
                 let type_streams = items.iter().map(|item| {
                     type_space
                         .id_to_entry
@@ -451,23 +578,36 @@ impl TypeEntry {
                 quote! { ( #(#type_streams),* ) }
             }
 
-            TypeEntry::Unit => quote! { () },
-            TypeEntry::String => quote! { String },
-            TypeEntry::BuiltIn(name) | TypeEntry::Integral(name) | TypeEntry::Float(name) => {
-                syn::parse_str::<syn::TypePath>(name)
-                    .unwrap()
-                    .to_token_stream()
-            }
+            TypeEntryDetails::Unit => quote! { () },
+            TypeEntryDetails::String => quote! { String },
+            TypeEntryDetails::Boolean => quote! { bool },
+            TypeEntryDetails::BuiltIn(name)
+            | TypeEntryDetails::Integer(name)
+            | TypeEntryDetails::Float(name) => syn::parse_str::<syn::TypePath>(name)
+                .unwrap()
+                .to_token_stream(),
 
-            TypeEntry::Reference(_) => panic!("references should be resolved by now"),
+            TypeEntryDetails::Reference(_) => panic!("references should be resolved by now"),
         }
     }
 
-    pub(crate) fn type_parameter_ident(&self, type_space: &TypeSpace) -> TokenStream {
-        match self {
+    pub(crate) fn type_parameter_ident(
+        &self,
+        type_space: &TypeSpace,
+        lifetime_name: Option<&str>,
+    ) -> TokenStream {
+        let lifetime = lifetime_name.map(|s| {
+            vec![
+                TokenTree::from(Punct::new('\'', Spacing::Joint)),
+                TokenTree::from(format_ident!("{}", s)),
+            ]
+            .into_iter()
+            .collect::<TokenStream>()
+        });
+        match &self.details {
             // We special-case enums for which all variants are simple to let
             // them be passed as values rather than as references.
-            TypeEntry::Enum(TypeEntryEnum{ variants, .. })
+            TypeEntryDetails::Enum(TypeEntryEnum{ variants, .. })
                 // TODO we should probably cache this rather than iterating
                 // every time. We'll know it when the enum is constructed.
                 if variants
@@ -477,68 +617,70 @@ impl TypeEntry {
                 self.type_ident(type_space, true)
             }
 
-            TypeEntry::Enum(_)
-            | TypeEntry::Struct(_)
-            | TypeEntry::Newtype(_)
-            | TypeEntry::Array(_)
-            | TypeEntry::Map(_, _)
-            | TypeEntry::Set(_)
-            | TypeEntry::BuiltIn(_) => {
+            TypeEntryDetails::Enum(_)
+            | TypeEntryDetails::Struct(_)
+            | TypeEntryDetails::Newtype(_)
+            | TypeEntryDetails::Array(_)
+            | TypeEntryDetails::Map(_)
+            | TypeEntryDetails::Set(_)
+            | TypeEntryDetails::Box(_)
+            | TypeEntryDetails::BuiltIn(_) => {
                 let ident = self.type_ident(type_space, true);
                 quote! {
-                    &#ident
+                    & #lifetime #ident
                 }
             }
 
-            TypeEntry::Option(id) => {
+            TypeEntryDetails::Option(id) => {
                 let inner_ty = type_space
                     .id_to_entry
                     .get(id)
                     .expect("unresolved type id for option");
-                let inner_ident = inner_ty.type_parameter_ident(type_space);
+                let inner_ident = inner_ty.type_parameter_ident(type_space, lifetime_name);
 
                 // Flatten nested Option types. This would only happen if the
                 // schema encoded it; it's an odd construction.
-                match inner_ty {
-                    TypeEntry::Option(_) => inner_ident,
+                match &inner_ty.details {
+                    TypeEntryDetails::Option(_) => inner_ident,
                     _ => quote! { Option<#inner_ident> },
                 }
             }
-            TypeEntry::Tuple(items) => {
+            TypeEntryDetails::Tuple(items) => {
                 let type_streams = items.iter().map(|item| {
                     type_space
                         .id_to_entry
                         .get(item)
                         .expect("unresolved type id for tuple")
-                        .type_parameter_ident(type_space)
+                        .type_parameter_ident(type_space, lifetime_name)
                 });
 
                 quote! { ( #(#type_streams),* ) }
             }
 
-            TypeEntry::Unit | TypeEntry::Integral(_) | TypeEntry::Float(_) => {
+            TypeEntryDetails::Unit | TypeEntryDetails::Boolean|TypeEntryDetails::Integer(_) | TypeEntryDetails::Float(_) => {
                 self.type_ident(type_space, true)
             }
-            TypeEntry::String => quote! { &str },
+            TypeEntryDetails::String => quote! { & #lifetime str },
 
-            TypeEntry::Reference(_) => panic!("references should be resolved by now"),
+            TypeEntryDetails::Reference(_) => panic!("references should be resolved by now"),
         }
     }
 
     pub(crate) fn describe(&self) -> String {
-        match self {
-            TypeEntry::Enum(TypeEntryEnum { name, .. }) => format!("enum {}", name),
-            TypeEntry::Struct(TypeEntryStruct { name, .. }) => format!("struct {}", name),
-            TypeEntry::Newtype(TypeEntryNewtype { name, type_id, .. }) => {
+        match &self.details {
+            TypeEntryDetails::Enum(TypeEntryEnum { name, .. }) => format!("enum {}", name),
+            TypeEntryDetails::Struct(TypeEntryStruct { name, .. }) => format!("struct {}", name),
+            TypeEntryDetails::Newtype(TypeEntryNewtype { name, type_id, .. }) => {
                 format!("newtype {} {}", name, type_id.0)
             }
 
-            TypeEntry::Unit => "()".to_string(),
-            TypeEntry::Option(type_id) => format!("option {}", type_id.0),
-            TypeEntry::Array(type_id) => format!("array {}", type_id.0),
-            TypeEntry::Map(key_id, value_id) => format!("map {} {}", key_id.0, value_id.0),
-            TypeEntry::Set(type_id) => format!("set {}", type_id.0),
-            TypeEntry::Tuple(type_ids) => {
+            TypeEntryDetails::Unit => "()".to_string(),
+            TypeEntryDetails::Option(type_id) => format!("option {}", type_id.0),
+            TypeEntryDetails::Array(type_id) => format!("array {}", type_id.0),
+            TypeEntryDetails::Map(type_id) => format!("map {}", type_id.0),
+            TypeEntryDetails::Set(type_id) => format!("set {}", type_id.0),
+            TypeEntryDetails::Box(type_id) => format!("box {}", type_id.0),
+            TypeEntryDetails::Tuple(type_ids) => {
                 format!(
                     "tuple ({})",
                     type_ids
@@ -548,19 +690,23 @@ impl TypeEntry {
                         .join(", ")
                 )
             }
-            TypeEntry::BuiltIn(name) | TypeEntry::Integral(name) | TypeEntry::Float(name) => {
-                name.clone()
-            }
-            TypeEntry::String => "string".to_string(),
+            TypeEntryDetails::Boolean => "bool".to_string(),
+            TypeEntryDetails::BuiltIn(name)
+            | TypeEntryDetails::Integer(name)
+            | TypeEntryDetails::Float(name) => name.clone(),
+            TypeEntryDetails::String => "string".to_string(),
 
-            TypeEntry::Reference(_) => unreachable!(),
+            TypeEntryDetails::Reference(_) => unreachable!(),
         }
     }
 }
 
 #[cfg(test)]
 mod tests {
-    use crate::{type_entry::TypeEntryStruct, TypeEntry, TypeSpace};
+    use crate::{
+        type_entry::{TypeEntry, TypeEntryStruct},
+        TypeEntryDetails, TypeSpace,
+    };
 
     #[test]
     fn test_ident() {
@@ -568,33 +714,38 @@ mod tests {
 
         let t = TypeEntry::new_integer("u32");
         let ident = t.type_ident(&ts, true);
-        let parameter = t.type_parameter_ident(&ts);
         assert_eq!(ident.to_string(), "u32");
+        let parameter = t.type_parameter_ident(&ts, None);
         assert_eq!(parameter.to_string(), "u32");
 
-        let t = TypeEntry::String;
+        let t = TypeEntry::from(TypeEntryDetails::String);
         let ident = t.type_ident(&ts, true);
-        let parameter = t.type_parameter_ident(&ts);
         assert_eq!(ident.to_string(), "String");
+        let parameter = t.type_parameter_ident(&ts, None);
         assert_eq!(parameter.to_string(), "& str");
+        let parameter = t.type_parameter_ident(&ts, Some("static"));
+        assert_eq!(parameter.to_string(), "& 'static str");
 
-        let t = TypeEntry::Unit;
+        let t = TypeEntry::from(TypeEntryDetails::Unit);
         let ident = t.type_ident(&ts, true);
-        let parameter = t.type_parameter_ident(&ts);
         assert_eq!(ident.to_string(), "()");
+        let parameter = t.type_parameter_ident(&ts, None);
         assert_eq!(parameter.to_string(), "()");
 
-        let t = TypeEntry::Struct(TypeEntryStruct {
+        let t = TypeEntry::from(TypeEntryDetails::Struct(TypeEntryStruct {
             name: "SomeType".to_string(),
             rename: None,
             description: None,
+            default: None,
             properties: vec![],
             deny_unknown_fields: false,
-        });
+        }));
 
         let ident = t.type_ident(&ts, true);
-        let parameter = t.type_parameter_ident(&ts);
         assert_eq!(ident.to_string(), "SomeType");
+        let parameter = t.type_parameter_ident(&ts, None);
         assert_eq!(parameter.to_string(), "& SomeType");
+        let parameter = t.type_parameter_ident(&ts, Some("a"));
+        assert_eq!(parameter.to_string(), "& 'a SomeType");
     }
 }
