@@ -1,12 +1,12 @@
-// Copyright 2021 Oxide Computer Company
+// Copyright 2022 Oxide Computer Company
 
-use std::collections::BTreeSet;
+use std::collections::HashSet;
 
-use convert_case::{Case, Casing};
 use schemars::schema::{
     ArrayValidation, InstanceType, Metadata, ObjectValidation, Schema, SchemaObject, SingleOrVec,
     SubschemaValidation,
 };
+use unicode_xid::UnicodeXID;
 
 use crate::Name;
 
@@ -175,91 +175,59 @@ fn schemas_mutually_exclusive(a: &Schema, b: &Schema) -> bool {
     }
 }
 
+// See if there are unique, required properties of each that cannot be present
+// in the other. In other words, see if there are properties that would
+// uniquely identify an objects as validating exclusively with one or the other
+// (but not with both).
 fn object_schemas_mutually_exclusive(
     a_validation: &ObjectValidation,
     b_validation: &ObjectValidation,
 ) -> bool {
-    match (a_validation, b_validation) {
-        (
-            ObjectValidation {
-                required: a_required,
-                additional_properties: Some(a_additional),
-                ..
-            },
-            ObjectValidation {
-                required: b_required,
-                additional_properties: Some(b_additional),
-                ..
-            },
-            // Both objects must disallow additional fields
-            // (this is stricter than absolutely necessary,
-            // but is a reasonable simplification)
-        ) if a_additional.as_ref() == &Schema::Bool(false)
-            && b_additional.as_ref() == &Schema::Bool(false) =>
-        {
-            // TODO seems odd that this matches
+    let ObjectValidation {
+        required: a_required,
+        properties: a_properties,
+        ..
+    } = a_validation;
+    let ObjectValidation {
+        required: b_required,
+        properties: b_properties,
+        ..
+    } = b_validation;
 
-            // Neither set of required properties must be a
-            // subset of the other i.e. both must have
-            // unique, required properties.
-            !a_required.is_subset(b_required) && !b_required.is_subset(a_required)
-        }
-        (
-            ObjectValidation {
-                properties: a_properties,
-                required: a_required,
-                // TODO I believe this should actually be
-                // Bool(false) as above
-                additional_properties: None,
-                ..
-            },
-            ObjectValidation {
-                properties: b_properties,
-                required: b_required,
-                additional_properties: None,
-                ..
-            },
-        ) => {
-            // The object schemas are mutually exclusive if
-            // each has one or more unique, required
-            // properties.
-            if !a_required.is_subset(b_required) && !b_required.is_subset(a_required) {
-                true
-            } else {
-                // Even if one of the sets of required
-                // properties is a subset of the other (or
-                // if they are identical), each may have
-                // properties that have fixed values that
-                // differ. This can happen in particular
-                // for internally or adjacently tagged
-                // enums where the properties may be
-                // identical but the value of the tag
-                // property will be unique.
+    // Either set of required properties must not be a subset of the other's
+    // properties i.e. if there's a property that *must* be in one of the two
+    // objects, and *cannot* be in the other, a property whose presence or
+    // absence determines which of the two objects is relevant.
+    if !a_required.is_subset(&b_properties.keys().cloned().collect())
+        || !b_required.is_subset(&a_properties.keys().cloned().collect())
+    {
+        true
+    } else {
+        // Even if all required properties of each is a permitted property of
+        // the other, each may have required properties that have fixed values
+        // that differ. This can happen in particular for internally or
+        // adjacently tagged enums where the properties may be identical but
+        // the value of the tag property will be unique.
 
-                // Compute the set that consists of
-                // fixed-value properties--a tuple of
-                // the property name and the value.
-                let aa = a_required
-                    .iter()
-                    .filter_map(|name| {
-                        let t = a_properties.get(name).unwrap();
-                        constant_string_value(t).map(|s| (name.clone(), s))
-                    })
-                    .collect::<BTreeSet<_>>();
-                let bb = b_required
-                    .iter()
-                    .filter_map(|name| {
-                        let t = b_properties.get(name).unwrap();
-                        constant_string_value(t).map(|s| (name.clone(), s))
-                    })
-                    .collect::<BTreeSet<_>>();
+        // Compute the set that consists of fixed-value properties--a
+        // tuple of the property name and the value.
+        let aa = a_required
+            .iter()
+            .filter_map(|name| {
+                let t = a_properties.get(name).unwrap();
+                constant_string_value(t).map(|s| (name.clone(), s))
+            })
+            .collect::<HashSet<_>>();
+        let bb = b_required
+            .iter()
+            .filter_map(|name| {
+                let t = b_properties.get(name).unwrap();
+                constant_string_value(t).map(|s| (name.clone(), s))
+            })
+            .collect::<HashSet<_>>();
 
-                // True if neither is a subset of the other.
-                !aa.is_subset(&bb) && !bb.is_subset(&aa)
-            }
-        }
-
-        _ => todo!(),
+        // True if neither is a subset of the other.
+        !aa.is_subset(&bb) && !bb.is_subset(&aa)
     }
 }
 
@@ -332,13 +300,33 @@ fn array_schemas_mutually_exclusive(
             a_max_items != b_max_items
         }
 
+        // Plain, vanilla arrays.
+        (
+            ArrayValidation {
+                items: Some(SingleOrVec::Single(a_items)),
+                additional_items: None,
+                max_items: None,
+                min_items: None,
+                unique_items: None,
+                contains: None,
+            },
+            ArrayValidation {
+                items: Some(SingleOrVec::Single(b_items)),
+                additional_items: None,
+                max_items: None,
+                min_items: None,
+                unique_items: None,
+                contains: None,
+            },
+        ) => schemas_mutually_exclusive(a_items, b_items),
+
         (aa, bb) => todo!("{:#?} {:#?}", aa, bb),
     }
 }
 
 /// If this schema represents a constant-value string, return that string,
 /// otherwise return None.
-pub(crate) fn constant_string_value(schema: &Schema) -> Option<String> {
+pub(crate) fn constant_string_value(schema: &Schema) -> Option<&str> {
     match schema {
         // Strings must be simple enumerations.
         Schema::Object(SchemaObject {
@@ -356,9 +344,7 @@ pub(crate) fn constant_string_value(schema: &Schema) -> Option<String> {
             extensions: _,
         }) if single.as_ref() == &InstanceType::String => {
             if values.len() == 1 {
-                values
-                    .get(0)
-                    .and_then(|value| value.as_str().map(ToString::to_string))
+                values.get(0).and_then(|value| value.as_str())
             } else {
                 None
             }
@@ -471,11 +457,37 @@ pub(crate) fn schema_is_named(schema: &Schema) -> Option<String> {
         _ => None,
     }?;
 
-    Some(sanitize(raw_name, Case::Pascal))
+    Some(sanitize(&raw_name, Case::Pascal))
 }
 
-fn sanitize(input: String, case: Case) -> String {
-    let out = input.replace("$", "-").replace("'", "").to_case(case);
+pub(crate) enum Case {
+    Pascal,
+    Snake,
+}
+
+pub(crate) fn sanitize(input: &str, case: Case) -> String {
+    use heck::{ToPascalCase, ToSnakeCase};
+    let to_case = match case {
+        Case::Pascal => str::to_pascal_case,
+        Case::Snake => str::to_snake_case,
+    };
+
+    // If every case was special then none of them would be.
+    let out = match input {
+        "+1" => "plus1".to_string(),
+        "-1" => "minus1".to_string(),
+        _ => to_case(
+            &input
+                .replace("'", "")
+                .replace(|c: char| !c.is_xid_continue(), "-"),
+        ),
+    };
+
+    let out = match out.chars().next() {
+        None => to_case("x"),
+        Some(c) if c.is_xid_start() => out,
+        Some(_) => format!("_{}", out),
+    };
 
     // Make sure the string is a valid Rust identifier.
     if syn::parse_str::<syn::Ident>(&out).is_ok() {
@@ -485,18 +497,17 @@ fn sanitize(input: String, case: Case) -> String {
     }
 }
 
-pub(crate) fn recase(input: String, case: Case) -> (String, Option<String>) {
-    let new = sanitize(input.clone(), case);
-    let rename = if new == input { None } else { Some(input) };
-    assert_ne!(new, "ref");
+pub(crate) fn recase(input: &str, case: Case) -> (String, Option<String>) {
+    let new = sanitize(&input, case);
+    let rename = if new == input {
+        None
+    } else {
+        Some(input.to_string())
+    };
     (new, rename)
 }
 
-pub(crate) fn get_type_name(
-    type_name: &Name,
-    metadata: &Option<Box<Metadata>>,
-    case: Case,
-) -> Option<String> {
+pub(crate) fn get_type_name(type_name: &Name, metadata: &Option<Box<Metadata>>) -> Option<String> {
     let name = match (type_name, metadata_title(metadata)) {
         (Name::Required(name), _) => name.clone(),
         (Name::Suggested(name), None) => name.clone(),
@@ -504,5 +515,133 @@ pub(crate) fn get_type_name(
         (Name::Unknown, None) => None?,
     };
 
-    Some(sanitize(name, case))
+    Some(sanitize(&name, Case::Pascal))
+}
+
+#[cfg(test)]
+mod tests {
+    use schemars::{schema_for, JsonSchema};
+
+    use crate::util::{sanitize, schemas_mutually_exclusive, Case};
+
+    #[test]
+    fn test_non_exclusive_structs() {
+        #![allow(dead_code)]
+
+        #[derive(JsonSchema)]
+        struct A {
+            a: Option<()>,
+            b: (),
+        }
+
+        #[derive(JsonSchema)]
+        struct B {
+            a: (),
+            b: Option<()>,
+        }
+
+        let a = schema_for!(A).schema.into();
+        let b = schema_for!(B).schema.into();
+
+        assert!(!schemas_mutually_exclusive(&a, &b));
+        assert!(!schemas_mutually_exclusive(&b, &a));
+    }
+
+    #[test]
+    fn test_unique_prop_structs() {
+        #![allow(dead_code)]
+
+        #[derive(JsonSchema)]
+        struct A {
+            a: Option<()>,
+            b: (),
+        }
+
+        #[derive(JsonSchema)]
+        struct B {
+            a: (),
+            b: Option<()>,
+            c: (),
+        }
+
+        let a = schema_for!(A).schema.into();
+        let b = schema_for!(B).schema.into();
+
+        assert!(schemas_mutually_exclusive(&a, &b));
+        assert!(schemas_mutually_exclusive(&b, &a));
+    }
+
+    #[test]
+    fn test_exclusive_structs() {
+        #![allow(dead_code)]
+
+        #[derive(JsonSchema)]
+        struct A {
+            a: Option<()>,
+            b: (),
+            aa: (),
+        }
+
+        #[derive(JsonSchema)]
+        struct B {
+            a: (),
+            b: Option<()>,
+            bb: (),
+        }
+
+        let a = schema_for!(A).schema.into();
+        let b = schema_for!(B).schema.into();
+
+        assert!(schemas_mutually_exclusive(&a, &b));
+        assert!(schemas_mutually_exclusive(&b, &a));
+    }
+
+    #[test]
+    fn test_exclusive_one_empty_struct() {
+        #![allow(dead_code)]
+
+        #[derive(JsonSchema)]
+        struct A {}
+
+        #[derive(JsonSchema)]
+        struct B {
+            a: (),
+            b: Option<()>,
+            bb: (),
+        }
+
+        let a = schema_for!(A).schema.into();
+        let b = schema_for!(B).schema.into();
+
+        assert!(schemas_mutually_exclusive(&a, &b));
+        assert!(schemas_mutually_exclusive(&b, &a));
+    }
+
+    #[test]
+    fn test_exclusive_simple_arrays() {
+        let a = schema_for!(Vec<u32>).schema.into();
+        let b = schema_for!(Vec<f32>).schema.into();
+
+        assert!(schemas_mutually_exclusive(&a, &b));
+        assert!(schemas_mutually_exclusive(&b, &a));
+    }
+
+    #[test]
+    fn test_sanitize() {
+        assert_eq!(sanitize("type", Case::Snake), "type_");
+        assert_eq!(sanitize("ref", Case::Snake), "ref_");
+        assert_eq!(sanitize("+1", Case::Snake), "plus1");
+        assert_eq!(sanitize("-1", Case::Snake), "minus1");
+        assert_eq!(sanitize("@timestamp", Case::Pascal), "Timestamp");
+        assert_eq!(sanitize("won't and can't", Case::Pascal), "WontAndCant");
+        assert_eq!(
+            sanitize(
+                "urn:ietf:params:scim:schemas:extension:gluu:2.0:user_",
+                Case::Pascal
+            ),
+            "UrnIetfParamsScimSchemasExtensionGluu20User"
+        );
+        assert_eq!(sanitize("Ipv6Net", Case::Snake), "ipv6_net");
+        assert_eq!(sanitize("V6", Case::Pascal), "V6");
+    }
 }
